/*
 * Author: Tony Rogvall <tony@bluetail.com>
 *
 * X11 Keyboard driver
 */
#include <unistd.h>
#include <errno.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include "device.h"

static int  X11_Open(KBDDEVICE *pkd);
static void X11_Close(void);
static void X11_GetModifierInfo(int *modifiers);
static int  X11_Read(MWUCHAR *buf, int *modifiers);

extern Display*     x11_dpy;
extern int          x11_scr;
extern Visual*      x11_vis;
extern Window       x11_win;
extern GC           x11_gc;
extern int          x11_setup_display();


KBDDEVICE kbddev = {
    X11_Open,
    X11_Close,
    X11_GetModifierInfo,
    X11_Read,
    NULL
};

/*
 * Open the keyboard.
 * This is real simple, we just use a special file handle
 * that allows non-blocking I/O, and put the terminal into
 * character mode.
 */
static int
X11_Open(KBDDEVICE *pkd)
{
    if (x11_setup_display() < 0)
	return -1;
    /* return the x11 file descriptor for select */
    return ConnectionNumber(x11_dpy);  
}

/*
 * Close the keyboard.
 * This resets the terminal modes.
 */
static void
X11_Close(void)
{
    /* nop */
}

/*
 * Return the possible modifiers for the keyboard.
 */
static  void
X11_GetModifierInfo(int *modifiers)
{
    *modifiers = 0;			/* no modifiers available */
}

/*
 * This reads one keystroke from the keyboard, and the current state of
 * the mode keys (ALT, SHIFT, CTRL).  Returns -1 on error, 0 if no data
 * is ready, and 1 if data was read.  This is a non-blocking call.
 */

static int grabbed = 0;
static int x11_accel_num;
static int x11_accel_den;
static int x11_thres;

static int
X11_Read(MWUCHAR *buf, int *modifiers)
{
    XEvent ev;

    /* check if we have a KeyPressedEvent */
    if (XCheckMaskEvent(x11_dpy, KeyPressMask, &ev)) {
	KeySym sym = XKeycodeToKeysym(x11_dpy, ev.xkey.keycode, 0);
	*modifiers = 0;         /* no modifiers yet */
	if (sym == NoSymbol)
	    return -1;
	else if (sym == XK_Escape) {
	    if (ev.xkey.state & ControlMask) { 
		/* toggle grab control */
		if (grabbed) {
		    XUngrabPointer(x11_dpy, CurrentTime);
		    XUngrabKeyboard(x11_dpy, CurrentTime);
		    XChangePointerControl(x11_dpy,
					  True, False,
					  x11_accel_num, 
					  x11_accel_den, 0);
		    grabbed = 0;
		}
		else {
		    /* save pointer config */
		    XGetPointerControl(x11_dpy,
				       &x11_accel_num,
				       &x11_accel_den,
				       &x11_thres);
		    XChangePointerControl(x11_dpy,
					  True, False,
					  1, 1, 0);
		    XGrabKeyboard(x11_dpy, 
				  x11_win,
				  True,  /* only to this window */
				  GrabModeAsync,
				  GrabModeAsync,
				  CurrentTime);
		    XGrabPointer(x11_dpy,
				 x11_win,
				 False,
				 PointerMotionMask | ButtonPressMask,
				 GrabModeAsync,
				 GrabModeAsync,
				 None,
				 None,
				 CurrentTime);
		    grabbed = 1;
		}
		return -1;
	    }
	    else if (grabbed)
		XChangePointerControl(x11_dpy,
				      True, False,
				      x11_accel_num, 
				      x11_accel_den, 0);
	    return -2;
	}
	else
	    *buf = sym & 0xff;  /* strip symbol to ascii */
	return 1;
    }
    return 0;
}
